from prompt_toolkit.completion import WordCompleter

cmds = {
    "keys": "pattern",
    "get": "key",
    "set": "key value [EX seconds] [PX milliseconds] [NX|XX]",
}

redis_completer = WordCompleter(cmds.keys(), meta_dict=cmds, ignore_case=True)



from redis import Redis, exceptions
from .result_processors import process_results, build_scoped_cmd


class RedisKeyMetadata:
    def __init__(self, kind, ttl):
        self.kind = kind
        self.ttl = ttl

    def __repr__(self):
        r = ""
        if self.kind:
            r += f"type: {self.kind}"
        if self.ttl:
            r += f"ttl: {self.ttl}"
        return r


class Executor:
    def __init__(self, host="localhost", port=6379, db=None):
        self.conn = Redis(host, port, db)

    def run(self, cmd, reset_cmd_stack=False):
        try:
            results = self.conn.execute_command(cmd)
            return cmd, process_results(cmd, results)
        except exceptions.ResponseError as e:
            return cmd, str(e)

    def scoped_run(self, item, prev_cmd):
        scoped_cmd = build_scoped_cmd(item, prev_cmd, self._get_data_type)
        if scoped_cmd is None:
            return None, None
        return self.run(scoped_cmd)

    def _get_data_type(self, key):
        try:
            results = self.conn.execute_command("type {}".format(key))
        except Exception as e:
            return None
        return results.decode("utf-8")

    def _get_ttl(self, key):
        try:
            ttl = self.con.execute_command("ttl {}".format(key))
        except Exception as e:
            return None
        return str(ttl)

    def get_metadata(self, key):
        kind = self._get_data_type(key)
        ttl = self._get_ttl(key)
        return RedisKeyMetadata(kind, ttl)



from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.filters import has_focus, Condition
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous


def create_key_bindings(trex_layout):
    @Condition
    def output_window_focused():
        _, split = trex_layout.active_split()
        return bool(split and trex_layout.layout.has_focus(split.content_window))

    @Condition
    def commandline_focused():
        _, split = trex_layout.active_split()
        return bool(split and trex_layout.layout.has_focus(split.commandline))

    redis_prompt = trex_layout.redis_prompt
    kb = KeyBindings()

    kb.add("tab")(focus_next)
    kb.add("s-tab")(focus_previous)

    @kb.add("c-d", filter=has_focus(redis_prompt))
    def _(event):
        if redis_prompt.text.strip() == "":
            event.app.exit()

    @kb.add("down", filter=has_focus(redis_prompt))
    @kb.add("c-j", filter=has_focus(redis_prompt))
    def _(event):
        output_splits = trex_layout.splits
        if output_splits:
            trex_layout.layout.focus(output_splits[-1])

    @kb.add("c-k", filter=output_window_focused)
    def _(event):
        trex_layout.layout.focus(redis_prompt)

    @kb.add("right", filter=output_window_focused)
    @kb.add("l", filter=output_window_focused)
    @kb.add("c-l", filter=output_window_focused)
    def _(event):
        output_splits = trex_layout.splits
        i, split = trex_layout.active_split()
        if (i + 1) < len(output_splits):
            trex_layout.layout.focus(output_splits[i + 1])

    @kb.add("left", filter=output_window_focused)
    @kb.add("h", filter=output_window_focused)
    @kb.add("c-h", filter=output_window_focused)
    def _(event):
        output_splits = trex_layout.splits
        i, split = trex_layout.active_split()
        if i > 0:
            trex_layout.layout.focus(output_splits[i - 1])

    @kb.add(":", filter=output_window_focused)
    def enter_command_mode(event):
        """
        Entering command mode.
        """
        _, active_split = trex_layout.active_split()
        cmd_buffer = active_split.commandline
        trex_layout.layout.focus(cmd_buffer.content)

    @kb.add("escape", filter=commandline_focused)
    @kb.add("c-c", filter=commandline_focused)
    def leave_command_mode(event):
        """
        Leaving command mode.
        """
        trex_layout.layout.focus_last()

    return kb



import re
import ast
from setuptools import setup, find_packages

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("t_rex/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

description = "Terminal Redis Explorer."

install_requirements = ["pygments", "prompt_toolkit>=2.0.0", "redis"]

setup(
    name="t_rex",
    version=version,
    license="BSD",
    packages=find_packages(),
    description=description,
    long_description=open("README.md").read(),
    install_requires=install_requirements,
    entry_points="""
        [console_scripts]
        t_rex=t_rex.main:run
        t-rex=t_rex.main:run
        trex=t_rex.main:run
    """,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



def process_results(cmd, output):
    try:
        processor = result_processors.get(_cmd_type(cmd))
        return processor and processor(output)
    except KeyError:
        return str(output)


def build_scoped_cmd(item, prev_cmd, get_data_type):
    cmd_type = _cmd_type(prev_cmd)
    cmd_builder = scoped_cmd_builders.get(cmd_type)
    if cmd_builder:
        return cmd_builder(item, prev_cmd, get_data_type)
    else:
        return None


def _extract_namespace(cmd):
    try:
        return cmd.split(None)[-1]
    except IndexError:
        return None


def _cmd_type(cmd):
    try:
        if cmd.startswith("config"):
            return " ".join(cmd.split(None, 2)[:2])
        else:
            return cmd.split(None, 1)[0].lower()
    except IndexError:
        return None


def _byteslist2strtuple(lst):
    """Converts a list into a pair to tuples where the subsequent items are
    paired together."""
    l = _bytes2str_list(lst)
    if len(l) < 2:
        return l
    return zip(l[0::2], l[1::2])


def _bytestuple2strtuple_list(lst):
    for tup in lst:
        if tup is None:
            yield "null"
        else:
            x, y = tup
            yield _bytes2str(x), _bytes2str(y)


def _bytes2str_list(lst):
    return [_bytes2str(x) for x in lst]


def _bytes2str(byte_s):
    if byte_s is None:
        return "null"
    return byte_s.decode("utf-8")


def _builder_HKEYS(item, prev_cmd, _):
    namespace = _extract_namespace(prev_cmd)
    return namespace and f"hget {namespace} {item}"


def _builder_KEYS(item, _, get_data_type):
    type2command = {
        "hash": "hkeys {item}",
        "string": "get {item}",
        "set": "smembers {item}",
        "list": "lrange {item} 0 -1",
        "zset": "zrange {item} 0 -1",
    }
    data_type = get_data_type(item)
    scoped_cmd = type2command.get(data_type)
    return scoped_cmd and scoped_cmd.format(item=item)


result_processors = {
    # connection
    "ping": _bytes2str,
    "echo": _bytes2str,
    # Geo
    "geoadd": str,
    "geohash": _bytes2str_list,
    "geopos": _bytestuple2strtuple_list,
    "geodist": _bytes2str,
    "georadius": _bytes2str_list,
    "georadiusbymember": _bytes2str_list,
    # hash
    "hdel": str,
    "hexists": _bytes2str,
    "hget": _bytes2str,
    "hgetall": _byteslist2strtuple,
    "hincrby": str,
    "hincrbyfloat": _bytes2str,
    "hkeys": _bytes2str_list,
    "hlen": str,
    "hmget": _bytes2str_list,
    "hmset": _bytes2str,
    "hset": str,
    "hsetnx": str,
    "hstrlen": str,
    "hvals": _bytes2str_list,
    "hscan": str,
    "keys": _bytes2str_list,
    "get": _bytes2str,
    "type": _bytes2str,
    "config get": _byteslist2strtuple,
    "config set": _bytes2str,
    "set": _bytes2str,
    "lpush": str,
    "lrange": _bytes2str_list,
    "sadd": str,
    "smembers": _bytes2str_list,
    "zadd": str,
    "zscore": _bytes2str,
    "zrange": _bytes2str_list,
    "zrangebyscore": _bytes2str_list,
}

scoped_cmd_builders = {"keys": _builder_KEYS, "hkeys": _builder_HKEYS}
